;
; Initialization and top-level loop for nametable editor
; Copyright 2012 Damian Yerrick
;
; Copying and distribution of this file, with or without
; modification, are permitted in any medium without royalty provided
; the copyright notice and this notice are preserved in all source
; code copies.  This file is offered as-is, without any warranty.
;

.include "src/nes.h"
.include "src/mmc1.h"
.include "src/ram.h"

.global OAM
.export nmi, reset, irq
.exportzp cur_keys, new_keys, das_keys, das_timer

.segment "ZEROPAGE"
nmis:          .res 1
oam_used:      .res 1  ; starts at 0
cur_keys:      .res 2
new_keys:      .res 2
das_keys:      .res 2
das_timer:     .res 2
tvSystem:      .res 1

.segment "CODE"
;;
; This NMI handler is good enough for a simple "has NMI occurred?"
; vblank-detect loop.  But sometimes there are things that you always
; want to happen every frame, even if the game logic takes far longer
; than usual.  These might include music or a scroll split.  In these
; cases, you'll need to put more logic into the NMI handler.
.proc nmi
  inc nmis
  rti
.endproc

; A null IRQ handler that just does RTI is useful to add breakpoints
; that survive a recompile.  Set your debugging emulator to trap on
; reads of $FFFE, and then you can BRK $00 whenever you need to add
; a breakpoint.
;
; But sometimes you'll want a non-null IRQ handler.
; On NROM, the IRQ handler is mostly used for the DMC IRQ, which was
; designed for gapless playback of sampled sounds but can also be
; (ab)used as a crude timer for a scroll split (e.g. status bar).
.proc irq
  rti
.endproc

; 
.proc reset
  ; The very first thing to do when powering on is to put all sources
  ; of interrupts into a known state.
  sei             ; Disable interrupts
  ldx #$00
  stx PPUCTRL     ; Disable NMI and set VRAM increment to 32
  stx PPUMASK     ; Disable rendering
  stx $4010       ; Disable DMC IRQ
  dex             ; Subtracting 1 from $00 gives $FF, which is a
  txs             ; quick way to set the stack pointer to $01FF
  bit PPUSTATUS   ; Acknowledge stray vblank NMI across reset
  bit SNDCHN      ; Acknowledge DMC IRQ
  lda #$40
  sta P2          ; Disable APU Frame IRQ
  lda #$0F
  sta SNDCHN      ; Disable DMC playback, initialize other channels

vwait1:
  bit PPUSTATUS   ; It takes one full frame for the PPU to become
  bpl vwait1      ; stable.  Wait for the first frame's vblank.

  ; We have about 29700 cycles to burn until the second frame's
  ; vblank.  Use this time to get most of the rest of the chipset
  ; into a known state.

  ; Most versions of the 6502 support a mode where ADC and SBC work
  ; with binary-coded decimal.  Some 6502-based platforms, such as
  ; Atari 2600, use this for scorekeeping.  The second-source 6502 in
  ; the NES ignores the mode setting because its decimal circuit is
  ; dummied out to save on patent royalties, and games either use
  ; software BCD routines or convert numbers to decimal every time
  ; they are displayed.  But some post-patent famiclones have a
  ; working decimal mode, so turn it off for best compatibility.
  cld

  ; Clear OAM and the zero page here.
  ; We don't copy the cleared OAM to the PPU until later.
  ldx #0
  jsr ppu_clear_oam  ; clear out OAM from X to end and set X to 0

  ; There are "holy wars" (perennial disagreements) on nesdev over
  ; whether it's appropriate to zero out RAM in the init code.  Some
  ; anti-zeroing people say it hides programming errors with reading
  ; uninitialized memory, and memory will need to be initialized
  ; again anyway at the start of each level.  Others in favor of
  ; clearing say that a lot more variables need set to 0 than to any
  ; other value, and a clear loop like this saves code size.  Still
  ; others point to the C language, whose specification requires that
  ; uninitialized variables be set to 0 before main() begins.
  txa
clear_zp:
  sta $00,x
  inx
  bne clear_zp
  
  ; Other things to do here (not shown):
  ; Set up PRG RAM
  ; Copy initial high scores, bankswitching trampolines, etc. to RAM
  ; Set up your sound engine
  
  lda #%00010
  ;        ^^ Vertical mirroring (horizontal arrangement of nametables)
  ;      ^^   32 KiB PRG switching
  ;     ^     8 KiB CHR switching
  jsr setMMC1BankMode

vwait2:
  bit PPUSTATUS  ; After the second vblank, we know the PPU has
  bpl vwait2     ; fully stabilized.
  
  lda #VBLANK_NMI
  sta PPUCTRL
  jsr getTVSystem
  sta tvSystem
  jmp play_game
.endproc


.proc play_game

  ; While in forced blank we have full access to VRAM.
  ; Copy CHR data to CHR RAM.
  jsr load_chr_ram

  ; Then load the nametable (background map).
  jsr title_screen
forever:
  jsr dispatch
  jmp forever
dispatch:
  lda screens+1,x
  pha
  lda screens,x
  pha
  rts

.pushseg
.segment "RODATA"
screens:
  .addr chrpicker-1, ntedit-1
  .addr drawing_help_screen-1, paledit-1, blank_unused_tiles-1
  .addr remap_all_dupes-1, defrag_tiles-1
  .addr comingsoon1_screen-1
.popseg
.endproc


BOXTILE_TOPLEFT = $10
BOXTILE_TOP = $11
BOXTILE_TOPRIGHT = $12
BOXTILE_LEFT = $13
BOXTILE_RIGHT = $14
BOXTILE_BOTTOMLEFT = $15
BOXTILE_BOTTOM = $16
BOXTILE_BOTTOMRIGHT = $17

.proc draw_box
left    = 0
top     = 1
width   = 2
height  = 3
addr_lo = 4
addr_hi = 5

  sec
  lda top
  .repeat 3
    ror a
    ror addr_lo
  .endrepeat
  sta PPUADDR
  sta addr_hi
  lda #%11100000
  and addr_lo
  ora left
  sta PPUADDR
  sta addr_lo
  
  ; Draw top of frame
  lda #0  ; turn off VRAM_DOWN
  sta PPUCTRL
  lda #BOXTILE_TOPLEFT
  sta PPUDATA
  lda #BOXTILE_TOP
  ldy width
toploop:
  sta PPUDATA
  dey
  bne toploop
  
  ; Draw right side of frame
  lda #VRAM_DOWN
  sta PPUCTRL
  lda #BOXTILE_TOPRIGHT
  sta PPUDATA
  lda #BOXTILE_RIGHT
  ldy height
rightloop:
  sta PPUDATA
  dey
  bne rightloop
  
  ; Draw left side of frame
  lda addr_hi
  sta PPUADDR
  lda addr_lo
  sta PPUADDR
  lda PPUDATA
  lda #BOXTILE_LEFT
  ldy height
leftloop:
  sta PPUDATA
  dey
  bne leftloop
  lda #0
  sta PPUCTRL
  lda #BOXTILE_BOTTOMLEFT
  sta PPUDATA
  lda #BOXTILE_BOTTOM
  ldy width
bottomloop:
  sta PPUDATA
  dey
  bne bottomloop
  lda #BOXTILE_BOTTOMRIGHT
  sta PPUDATA
  rts
.endproc

